
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#ifndef PG_BINARY_STREAM_H
#define PG_BINARY_STREAM_H

// perhaps operators for these should be moved into a separate file?
#include "Time.h"
#include "Size.h"
#include "NetDouble.h"
#include "BigSize.h"
#include "String.h"
#include "Array.h"
#include "NetAddr.h"

// a collection of streams that operate on binary objects
// storage format is platform-independent
// technically, each stream is a wrapper around corresponding
// iostream

// some environments do not know better than #define these
#ifdef getc
#undef getc
#endif
#ifdef putc
#undef putc
#endif

class istream;
class ostream;
class iostream;

// base class for all binary streams
class BinIos {
	public:
		virtual ~BinIos();

		const String name() const { return theName; }

	protected:
		String theName;

};

class IBStream: public virtual BinIos {
	public:
		typedef void (Manip)(class IBStream &);

	public:
		IBStream();
		~IBStream();

		void configure(istream *aStream, const String &aName);

		operator void*() const { return theStream && theStream->good() ? theStream : 0; }

		char getc() { char c; get(&c, 1); return c; }
		bool getb() { return (bool)getc(); }
		int geti() { int x; return(geti(x)); }
		int geti(int &x) { get(&x, SizeOf(x)); return x = ntohl(x); }
		String &gets(String &s);

	protected:
		void get(void *buf, int len) { theStream->read((char*)buf, len); }

	protected:
		istream *theStream;
};

class OBStream: public virtual BinIos {
	public:
		typedef void (Manip)(class OBStream &);

	public:
		OBStream();
		~OBStream();

		void configure(ostream *aStream, const String &aName);

		operator void*() const { return theStream && theStream->good() ? theStream : 0; }

		void putc(char c) { put(&c, 1); }
		void putb(bool b) { putc((char)(b ? 1 : 0)); }
		void puti(int x) { x = htonl(x); put(&x, SizeOf(x)); }
		void puts(const char *s, Size size) { puti(size); if (size) put(s, size); }

	protected:
		void put(const void *buf, Size size);

	protected:
		ostream *theStream;
};

class BStream: public IBStream, public OBStream {
	public:
		void configure(iostream *aStream, const String &aName);
};



/* operators for common types */

inline
IBStream &operator >>(IBStream &is, char &c) {
	c = is.getc();
	return is;
}

inline
IBStream &operator >>(IBStream &is, bool &b) {
	b = is.getb();
	return is;
}

inline
IBStream &operator >>(IBStream &is, int &x) {
	is.geti(x);
	return is;
}

inline
IBStream &operator >>(IBStream &is, double &x) {
	NetDouble nd;
	is >> nd.mnt >> nd.exp;
	x = nd;
	return is;
}

inline
IBStream &operator >>(IBStream &is, String &s) {
	is.gets(s);
	return is;
}

inline
IBStream &operator >>(IBStream &is, Time &t) {
	t.tv_sec = is.geti();
	t.tv_usec = is.geti();
	return is;
}

inline
IBStream &operator >>(IBStream &is, BigSize &bs) {
	return is >> bs.theAcc >> bs.theCnt;
}

inline
IBStream &operator >>(IBStream &is, Size &sz) {
	return is >> sz.theSize;
}

inline
IBStream &operator >>(IBStream &is, NetAddr &a) {
	in_addr n;
	n.s_addr = is.geti();
	const int port = is.geti();
	a = NetAddr(n, port);
	return is;
}

inline
IBStream &operator >>(IBStream &is, IBStream::Manip m) {
	m(is);
	return is; 
}

template <class Item>
inline
IBStream &operator >>(IBStream &is, Array<Item> &a) {
	const int cnt = is.geti();
	if (a.count()) { // use existing entries
		Assert(cnt <= a.count());
		for (int i = 0; is && i < cnt; ++i)
			is >> a.item(i);
	} else {
		for (int i = 0; is && i < cnt; ++i) {
			Item item;
			if (is >> item)
				a << item;
		}
	}
	return is;
}


inline
OBStream &operator <<(OBStream &os, char c) {
	os.putc(c);
	return os;
}

inline
OBStream &operator <<(OBStream &os, bool b) {
	os.putb(b);
	return os;
}

inline
OBStream &operator <<(OBStream &os, int x) {
	os.puti(x);
	return os;
}

inline 
OBStream &operator <<(OBStream &os, double x) { 
	NetDouble nd(x);
	return os << nd.mnt << nd.exp;
}

inline
OBStream &operator <<(OBStream &os, const char *s) {
	os.puts(s, strlen(s)+1);
	return os;
}

inline
OBStream &operator <<(OBStream &os, const String &s) {
	if (s)
		os.puts(s.str(), s.len()+1);
	else
		os.puts(0, 0);
	return os;
}

inline
OBStream &operator <<(OBStream &os, const Time &t) {
	os.puti(t.tv_sec);
	os.puti(t.tv_usec);
	return os;
}

inline
OBStream &operator <<(OBStream &os, const BigSize &bs) {
	return os << bs.theAcc << bs.theCnt;
}

// XXX: should log either A or N
inline
OBStream &operator <<(OBStream &os, const NetAddr &a) {
	return os << (int)a.addrN().s_addr << a.port();
}

inline
OBStream &operator <<(OBStream &os, OBStream::Manip m) {
	m(os);
	return os;
}

template <class Item>
inline
OBStream &operator <<(OBStream &os, const Array<Item> &a) {
	os.puti(a.count());
	for (int i = 0; i < a.count(); ++i)
		os << a[i];
	return os;
}

#endif
