
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "polygraph.h"

#include "base/OLog.h"
#include "runtime/LogComment.h"
#include "runtime/SharedOpts.h"
#include "runtime/StatIntvl.h"
#include "XactFarm.h"

#include "AsyncClt.h"
#include "SyncClt.h"
#include "CltOpts.h"
#include "ServerRep.h"

#include "Server.h"
#include "SrvOpts.h"

#include "Cache.h"
#include "DistrPoint.h"
#include "IcpProxy.h"
#include "PxyCltXact.h"
#include "PxySrvXact.h"
#include "PolyPxy.h"

#include "PglSemx.h"
#include "CacheSym.h"
#include "RobotSym.h"
#include "ServerSym.h"
#include "ProxySym.h"

#include "polyLogCats.h"
#include "gadgets.h"
#include "rndDistrs.h"
#include "runtime/globals.h"

typedef XactFarmT<CltXact, PxyCltXact> PxyCltXactFarm;
typedef XactFarmT<SrvXact, PxySrvXact> PxySrvXactFarm;


PolyPxy::PolyPxy() {
	theAgentType = "Proxy";
}

void PolyPxy::makeAgent(const AgentCfg &cfg) {
	Assert(cfg.sym);
	
	const ProxySym *pcfg = &(const ProxySym&)cfg.sym->cast("Proxy");
	int icpSrv = -1, icpClt = -1; // port numbers
	Client *client = 0;
	Server *server = 0;
	Cache *cache = 0;

	if (const CacheSym *ccfg = pcfg->cache()) {
		cache = new Cache;
		cache->configure(ccfg);
		ccfg->icpPort(icpSrv);
	} else {
		cerr << pcfg->loc() << "must configure cache for the proxy" << endl;
		exit(-3);
	}

	if (const RobotSym *rcfg = pcfg->client()) {
		if (RndDistr *iad = rcfg->reqInterArrival())
			client = new AsyncClt(iad);
		else
			client = new SyncClt();

		client->configure(rcfg, *cfg.host, agentMustBind);
		addAgent(client);
		rcfg->icpPort(icpClt);
	}

	if (const ServerSym *scfg = pcfg->server()) {
		server = new Server();
		server->configure(scfg, *cfg.host, agentMustBind);
		addAgent(server);
	}

	IcpClient *icpClient = 0;
	const in_addr host = cfg.host->addrN();
	if (icpClt == icpSrv && icpSrv >= 0) {
		IcpProxy *i = new IcpProxy;
		i->configure(NetAddr(host, icpSrv));
		i->cache(cache);
		addIcpAgent(i);
		icpClient = i;
	} else {
		if (icpSrv >= 0) {
			IcpServer *i = new IcpServer;
			i->configure(NetAddr(host, icpSrv));
			i->cache(cache);
			addIcpAgent(i);
		}
		if (icpClt >= 0) {
			IcpClient *i = new IcpClient;
			i->configure(NetAddr(host, icpClt));
			addIcpAgent(i);
			icpClient = i;
		}
	}

	if (client) {
		client->cache(cache);
		client->icpClient(icpClient);
	}

	if (server)
		server->cache(cache);

	cache->client(client);
}

void PolyPxy::addIcpAgent(IcpBase *icpAgent) {
	Assert(icpAgent);
	theIcpAgents.append(icpAgent);
}

void PolyPxy::configure() {
	PolyApp::configure();

	// new oid prefetching
	ServerRep::ConfigureNids(TheCltOpts.theNidsPrefDepth, TheOpts.theNidsPerMsgMax);

	Client::Farm(new PxyCltXactFarm);
	Server::Farm(new PxySrvXactFarm);

	IcpCltXact::TheTimeout = TheCltOpts.theIcpTout;
	StatIntvl::ActiveCat(lgcCltSide);
	StatIntvl::ActiveCat(lgcSrvSide);
	StatIntvl::TheReportCat = lgcCltSide;
}

void PolyPxy::getOpts(Array<OptGrp*> &opts) {
	PolyApp::getOpts(opts);
	opts.append(&TheCltOpts);
	opts.append(&TheSrvOpts);
}

void PolyPxy::reportCfg() {
	PolyApp::reportCfg();
}

void PolyPxy::logState(OLog &log) {
	PolyApp::logState(log);
	Client::LogState(log);
}

void PolyPxy::startAgents() {
	Comment(1) << "starting " << theIcpAgents.count() << " ICP agents..." << endc;
	for (int i = 0; i < theIcpAgents.count(); ++i)
		theIcpAgents[i]->start();
	PolyApp::startAgents();
}

void PolyPxy::step() {
	PolyApp::step();
	DistrPoint::CallAll();
}

int main(int argc, char *argv[]) {
	PolyPxy app;
	return app.run(argc, argv);
}
