
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#ifndef PG_HTTP_HEADERS_H
#define PG_HTTP_HEADERS_H

#include <iostream.h>

#include "Array.h"
#include "NetAddr.h"
#include "base/UniqId.h"
#include "base/Oid.h"
#include "LibInit.h"

class MsgHdrParsTab;

class MsgHdr {
	public:
		typedef bool (MsgHdr::*Parser)(const char *buf, const char *eoh);

	public:
		MsgHdr(const MsgHdrParsTab &aTab);
		virtual ~MsgHdr() {}

		virtual void reset();

		bool keepAlive() const;

		bool parse(const char *buf, Size sz);

		virtual void store(OLog &log) const;

	protected:
		static void Configure(MsgHdrParsTab &tab);
		static int AddParser(const String &field, Parser parser, MsgHdrParsTab &tab);

	protected:
		void parseFields();
		bool parseUniqId(UniqId &id, const char *&buf, const char *eoh);

		virtual bool parseRLine(const char *buf, const char *eol) = 0;

		// can be used by both requests and replies
		bool parseDate(const char *buf, const char *eoh);
		bool parseContLen(const char *buf, const char *eoh);
		bool parsePragma(const char *buf, const char *eoh);
		bool parseCControl(const char *buf, const char *eoh);
		bool parseXXact(const char *buf, const char *eoh);
		bool parseConnection(const char *buf, const char *eoh);

		// these have to be here so we can register them in req
		virtual bool parseHost(const char *buf, const char *eoh);
		virtual bool parseIms(const char *buf, const char *eoh);
		virtual bool parseXNidShortage(const char *buf, const char *eoh);
		virtual bool parseXFailedOid(const char *buf, const char *eoh);
		virtual bool parseXCltWss(const char *buf, const char *eoh);

		// these have to be here so we can register them in rep
		virtual bool parseLMT(const char *buf, const char *eoh);
		virtual bool parseExpires(const char *buf, const char *eoh);
		virtual bool parseXSrvWid(const char *buf, const char *eoh);
		virtual bool parseXSrvSets(const char *buf, const char *eoh);
		virtual bool parseXSrvNids(const char *buf, const char *eoh);

	public:
		const MsgHdrParsTab &theParsTab;

		Size theHdrSize;
		Time theDate;      // Date: header field
		Size theContSize;
		UniqId theGroupId;
		UniqId theXactId;
		int theSendPos;  // position of message being send
		int theRecvPos;  // the latest message received

		int theNidPack;  // new oids pack identifier
		int theNidStart; // new oids pack start
		int theNidCount; // new oids in a pack

		double theHttpVer;
		enum { kaNo, kaDefault, kaYes } theKeepAliveConn;
		bool isCachable;

	protected:
		Array<const char *> theFields; // positions of header fields
		const char *theBufBeg, *theBufEnd;
		const char *theSrchPtr;        // parse buffer iterator
		enum { ssFirst, ssSkip, ssFound } theSrchState;
};

class ReqHdr: public MsgHdr {
	public:
		static void Configure();
		static void Clean();
		
	public:
		ReqHdr();

		virtual void reset();

		bool foreign() const { return theOidRec.oid <= 0; }
		bool parseUrl(const char *&buf, const char *end);

		virtual void store(OLog &log) const;

	protected:
		virtual bool parseRLine(const char *buf, const char *eol);

		virtual bool parseHost(const char *buf, const char *eoh);
		virtual bool parseIms(const char *buf, const char *eoh);
		virtual bool parseXNidShortage(const char *buf, const char *eoh);
		virtual bool parseXFailedOid(const char *buf, const char *eoh);
		virtual bool parseXCltWss(const char *buf, const char *eoh);

		bool parseHostInURL(const char *&start, const char *eorl);

	protected:
		static MsgHdrParsTab *TheParsTab;

	public:
		OidRec theOidRec;
		NetAddr theHost;
		Time theIms;       // If-Modified-Since
		int theFailedOid;
		int theCltWss;     // clients idea of the working set size
};

class RepHdr: public MsgHdr {
	public:
		enum StatusCode { scUnknown = -1, scOK = 200, scNoContent = 204, scNotModified = 304 };

	public:
		static void Configure();
		static void Clean();
		
	public:
		RepHdr();

		virtual void reset();

		bool polyHeaders() const { return theXactId && theSrvWid; }
		bool foreign() const { return expectBody() && !polyHeaders(); }
		bool expectBody() const;
		Time calcLmt() const;

		virtual void store(OLog &log) const;

	protected:
		virtual bool parseRLine(const char *buf, const char *eol);

		virtual bool parseLMT(const char *buf, const char *eoh);
		virtual bool parseExpires(const char *buf, const char *eoh);
		virtual bool parseXSrvWid(const char *buf, const char *eoh);
		virtual bool parseXSrvSets(const char *buf, const char *eoh);
		virtual bool parseXSrvNids(const char *buf, const char *eoh);

	protected:
		static MsgHdrParsTab *TheParsTab;

	public:
		int theStatus;    // HTTP status code
		Time theLMT;      // Last-Modified-Time header field
		Time theExpires;  // Expires HTTP header field
		UniqId theSrvWid; // server world id
		int theSecOid;    // "secured" oid,
		int theSrvWss;    // server's idea of the working set size
		int theHotSetPos; // hot set position
};

inline ostream &crlf(ostream &os) { return os.write("\r\n", 2); }

LIB_INITIALIZER(HttpHdrsInit)

#endif
