
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "polygraph.h"

#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>

#ifdef HAVE_NETINET_IP_DUMMYNET_H

#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/ip_fw.h>
#include <net/route.h>
#include <sys/param.h>
#include <sys/mbuf.h>
#include <netinet/ip_dummynet.h>

#include "PglPp.h"
#include "PglParser.h"
#include "PglSemx.h"
#include "NetPipeSym.h"
#include "NetPipeArrIter.h"

#include "IpSocket.h"
#include "base/RndPermut.h"
#include "base/CmdLine.h"
#include "base/opts.h"
#include "getIfAddrs.h"
#include "gadgets.h"


class MyOpts: public OptGrp {
	public:
		MyOpts():
			theHelpOpt(this,     "help",           "list of options"),
			theCfgFileName(this, "config <file>",  "configuration file name"),
			theCfgDirs(this,     "cfg_dirs <list>","directories to search for cfg files"),
			theRngSeed(this,     "rng_seed <int>", "r.n.g. seed", 1)
			{}

		virtual bool validate() const;

		virtual ostream &printAnonym(ostream &os) const;
		virtual bool parseAnonym(const Array<const char *> &opts);
		virtual bool canParseAnonym() const { return true; }

	public:
		HelpOpt theHelpOpt;
		StrOpt theCfgFileName;
		StrArrOpt theCfgDirs;
		IntOpt theRngSeed;
};

struct NetPipe: public dn_pipe {
	NetPipe() { memset(this, 0, sizeof(*this)); }
};

struct IpFwRule: public ip_fw {
	IpFwRule() { memset(this, 0, sizeof(*this)); }
};

struct PipedAddr {
	in_addr addr;
	int pipeCount;
	PipedAddr(): pipeCount(0) { addr.s_addr = 0; }
	PipedAddr(in_addr a): addr(a), pipeCount(0) {}
};

static MyOpts TheOpts;
static Array<PipedAddr> ThePipedAddrs;
static int ThePipeCount;


/* MyOpt */

ostream &MyOpts::printAnonym(ostream &os) const {
	return os << "[config_file_name]";
}

bool MyOpts::parseAnonym(const Array<const char *> &opts) {
	if (opts.count() < 1)
		return true;

	if (opts.count() > 1)
		return false;

	if (theCfgFileName) {
		cerr << "cannot specify configuration file twice" << endl;
		return false;
	}

	theCfgFileName.val(opts[0]);
	return true;
}

bool MyOpts::validate() const {
	if (!theCfgFileName)
		cerr << "configuration file must be specified" << endl;
	else
		return OptGrp::validate();
	return false;
}

static
void parseConfigFile(const String &fname) {
	Assert(fname);
	TheOpts.theCfgDirs.copy(PglPp::TheDirs);

	PglPp pp(fname);
	PglParser parser(&pp);

	parser.parse();
}

static
void configureLogs(int prec) {
	configureStream(cout, prec);
	configureStream(cerr, prec);
	configureStream(clog, prec);
}

static
void configure() {
	configureLogs(2);
	parseConfigFile(TheOpts.theCfgFileName);

	if (!PglSemx::TheNetPipesToUse.count()) {
		cerr << "no pipe specs found in " 
			<< TheOpts.theCfgFileName << ", exiting" << endl;
		exit(0);
	}

	// set random seeds
	RndPermutSeed(TheOpts.theRngSeed);
	RndGen::TheDefSeed = RndPermut(TheOpts.theRngSeed);
}

static
bool ruleBelongs(const ip_fw &rule, const Array<in_addr> &addrs) {
	for (int i = 0; i < addrs.count(); ++i) {
		const in_addr &addr = addrs[i];
		if (rule.fw_src == addr || rule.fw_dst == addr)
			return true;
	}
	return false;
}

static
void delPipes(const Array<in_addr> &addrs) {
	IpSocket s;
	Must(s.create());

	Array<ip_fw> rules;
	Must(s.getIpFwRules(rules));

	for(int i = 0; i < rules.count(); ++i) {
		const ip_fw &rule = rules[i];
		if (ruleBelongs(rule, addrs)) {
			if (rule.fw_flg & IP_FW_F_PIPE) {
				NetPipe pipe;
				pipe.pipe_nr = rule.fw_un.fu_pipe_nr;
				Should(s.delDnPipe(pipe));
			}
			Should(s.delIpFwRule(rule));
		}
	}

	s.close();
}

// returns next 'pipe number' for an address
// increments number of pipes for that address
static
int nextPipeNum(const NetAddr &addr) {
	ThePipeCount++;

	for (int i = 0; i < ThePipedAddrs.count(); ++i) {
		if (ThePipedAddrs[i].addr == addr.addrN())
			return ThePipeCount;
	}

	// new address
	ThePipedAddrs.append(PipedAddr(addr.addrN()));
	return ThePipeCount;
}

static
in_addr addressMask(const NetAddr &) {
	in_addr mask;
	mask.s_addr = INADDR_BROADCAST;
	return mask;
}

static
void addPipe(const NetPipeSym *ps, const NetAddr &addr, Time delay, bool incoming) {

	// raw IP socket
	IpSocket s;
	Must(s.create());
	
	/* create a pipe */

	NetPipe pipe;
	pipe.pipe_nr = nextPipeNum(addr);
	ps->bandwidthBitsps(pipe.bandwidth);
	ps->queuePacketSize(pipe.queue_size);
	ps->queueByteSize(pipe.queue_size_bytes);
	if (delay > 0)
		pipe.delay = delay.msec();
	double plr = 0;
	if (ps->packetLoss(plr))
		pipe.plr = (int)MiniMax(0.0, INT_MAX*plr, (double)INT_MAX);

	Must(s.configureDnPipe(pipe));

	/* add pipe to the ipfw rule */

	IpFwRule rule;
	rule.fw_number = pipe.pipe_nr;
	rule.fw_flg = IP_FW_F_PIPE;
	rule.fw_divert_port = pipe.pipe_nr;
	rule.fw_prot = IPPROTO_IP;

	if (incoming) {
		rule.fw_dst = addr.addrN();
		rule.fw_dmsk = addressMask(addr);
		rule.fw_flg |= IP_FW_F_IN;
	} else {
		rule.fw_src = addr.addrN();
		rule.fw_smsk = addressMask(addr);
		rule.fw_flg |= IP_FW_F_OUT;
	}
	
	Must(s.addIpFwRule(rule));

	s.close();
}

static
void addPipe(const NetPipeSym *ps, const NetAddr &addr) {
	// if PGL pipe does not define a direction, both directions are used
	// and the delay should be the same for both directions

	Time delay;
	if (RndDistr *d = ps->packetDelay()) {
		// custom rng, to avoid repeated sequences when many ps:s are used
		static RndGen rng;
		d->rndGen(&rng);
		delay = Time::Secd(d->trial());
	}

	if (ps->incoming())
		addPipe(ps, addr, delay, true);
	if (ps->outgoing())
		addPipe(ps, addr, delay, false);
}

static
void run() {
	Array<in_addr> localAddrs;
	Must(GetIfAddrs(localAddrs));

	clog << "detected " << localAddrs.count() << " local addresses" << endl;

	// delete old pipes
	delPipes(localAddrs);

	// add new pipes
	for (NetPipeArrIter i(PglSemx::TheNetPipesToUse, &localAddrs); i; ++i) {
		NetPipeSym *ps= i.pipeSym();
		const NetAddr &host = i.host();
		addPipe(ps, host);
	}

	if (!ThePipedAddrs.count()) {
		cerr << "no pipes matched local addresses" << endl;

		cerr << "pipe addresses:";
		for (NetPipeArrIter i(PglSemx::TheNetPipesToUse); i; ++i)
			cerr << ' ' << i.host();
		cerr << endl;

		cerr << "local addresses: ";
		for (int i = 0; i < localAddrs.count(); ++i)
			cerr << ' ' << inet_ntoa(localAddrs[i]);
		cerr << endl;

		exit(-2);
	}

	clog << "added " << ThePipeCount << " pipes." << endl;
	clog << "configured " << ThePipedAddrs.count() << " addresses." << endl;
}

int main(int argc, char *argv[]) {

	CmdLine cmd;
	cmd.configure(Array<OptGrp*>() << &TheOpts);
	if (!cmd.parse(argc, argv) || !TheOpts.validate())
		return -1;

	configure();
	run();
	return 0;
}

#else

#warning piper cannot be built without a properly installed DummyNet kernel feature
#warning piper is a configuration tool that is not required to run benchmarks
int main(int, char *argv[]) {
	cerr << argv[0] << ": DummyNet must be [correctly] installed to use this tool." << endl;
	return -1;
}

#endif /* HAVE_NETINET_IP_DUMMYNET_H */
